import builtins
from importlib.util import find_spec
import os
import sys
import types


def noop(*args):
    """A no-op function.

    Args:
        *args: The arguments to ignore
    """
    pass


original_import = builtins.__import__


def get_custom_import(modified_file, new_source_code_file):
    """Get a custom import function that disables reflex_ai.enable,
    and replaces imports of the modified file the code living in new_source_code_file.

    Args:
        modified_file: The name of the modified file.
        new_source_code_file: The new source code file where the changes are made.

    Returns:
        The custom import function.
    """

    def custom_import(name, globals=None, locals=None, fromlist=(), level=0) -> object:
        """A custom import to disable reflex_ai.enable.

        Args:
            name: The name of the module to import.
            globals: The global namespace.
            locals: The local namespace.
            fromlist: The list of names to import.
            level: The level in the package hierarchy.

        Returns:
            The imported module.
        """

        # Try to find a spec for this name we are importing.
        # Note: Sometimes this will raise an exception, generally we don't care why (we only care if a spec exists, and it corresponds to the file we modified).
        try:
            spec = find_spec(name)
        except Exception:
            spec = None

        if spec is not None and spec.origin == modified_file:
            new_module = types.ModuleType(name)
            with open(new_source_code_file, "r") as f:
                new_source = f.read()
            exec(new_source, new_module.__dict__)
            return new_module

        module = original_import(name, globals, locals, fromlist, level)

        # Disable reflex_ai.enable
        if name == "reflex_ai" and fromlist and "enable" in fromlist:
            module.enable = noop
        return module

    return custom_import


def validate_source(
    new_source_code_file: str,
    modified_file: str,
    validation_function: str,
    validation_file: str,
):
    """Validate the source code by running the validation function.

    Args:
        new_source_code_file: The new source code file where the changes are made.
        modified_file: The name of the modified file.
        validation_function: The validation function to run.
        validation_file: The file containing the validation function.

    Raises:
        Exception: If the code being validated raises an exception.
    """
    builtins.__import__ = get_custom_import(modified_file, new_source_code_file)
    try:
        # Read the source code
        file_with_source_to_exec = (
            new_source_code_file
            if modified_file == validation_file
            else validation_file
        )

        with open(file_with_source_to_exec) as f:
            source_to_exec = f.read()

        # Add the current directory to the path.
        if os.getcwd() not in sys.path:
            sys.path.append(os.getcwd())

        env = {}
        try:
            # Execute the source code.
            exec(source_to_exec, env, env)

            # Run the validation function.
            component_fn = env[validation_function]

            eval(f"{component_fn.__name__}()", env, env)

        except Exception as e:
            print(str(e), file=sys.stderr)
            raise

    finally:
        builtins.__import__ = original_import


if __name__ == "__main__":
    validate_source(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4])



import reflex as rx
from reflex_ai.toolbar import ToolbarState
from reflex_ai.components import hint, icon_button


def help_icon() -> rx.Component:
    def popover_item(text: str, icon: str, link: str = "/") -> rx.Component:
        return rx.box(
            rx.link(
                rx.icon(tag=icon, size=20, stroke_width="1.5"),
                text,
                href=link,
                underline="none",
                class_name="flex items-center gap-3 hover:bg-[--slate-3] px-[0.875rem] py-2 font-base text-[--slate-10] hover:!text-[--slate-9] transition-bg overflow-hidden",
            ),
        )

    return rx.popover.root(
        rx.popover.trigger(
            rx.box(
                hint(
                    text="Support",
                    content=rx.box(
                        rx.icon(
                            tag="circle-help",
                            size=22,
                            stroke_width="1.5",
                            class_name="!text-[--slate-10]",
                        ),
                        class_name="group-data-[state=open]:bg-[--slate-3] group-hover:bg-[--slate-3] p-2 rounded-xl transition-bg cursor-help",
                    ),
                ),
                class_name="group",
            ),
        ),
        rx.popover.content(
            rx.box(
                popover_item(
                    "Documentation", "book-text", "https://reflex.dev/"
                ),  # TODO: add docs
                popover_item(
                    "Support", "life-buoy", "https://discord.com/invite/T5WSbC2YtQ"
                ),
                class_name="border-[--slate-5] bg-[--slate-1] shadow-large border border-box rounded-xl divide-y divide-[var(--slate-5)] overflow-hidden",
            ),
            side="top",
            avoid_collisions=True,
            class_name="items-center bg-transparent !shadow-none !p-0 border-none w-[214px] overflow-visible",
        ),
    )


def color_item(color: str) -> rx.Component:
    from reflex_ai.playground import SettingsState

    return rx.box(
        rx.box(
            rx.cond(
                SettingsState.color == color,
                rx.icon(
                    tag="check", size=13, stroke_width="2.5", class_name="!text-white"
                ),
                rx.fragment(),
            ),
            class_name="w-5 h-5 rounded-full flex items-center justify-center",
            style={
                "background": f"var(--{color}-9)",
            },
        ),
        rx.text(color.capitalize(), class_name="text-[--slate-12] font-small"),
        class_name="flex items-center gap-2 rounded-md shadow-small px-3 cursor-pointer bg-[--slate-1] hover:bg-[--slate-3] transition-bg box-border h-8",
        outline=rx.cond(
            SettingsState.color == color,
            "2px solid var(--slate-12)",
            "1px solid var(--slate-6)",
        ),
        on_click=SettingsState.set_color(color),
    )


def settings_icon() -> rx.Component:
    from reflex_ai.playground import SettingsState

    colors = ["violet", "yellow", "green", "blue", "orange", "red"]

    return rx.popover.root(
        rx.popover.trigger(
            rx.box(
                hint(
                    text="Settings",
                    content=rx.box(
                        rx.icon(
                            tag="cog",
                            size=22,
                            stroke_width="1.5",
                            class_name="!text-[--slate-10]",
                        ),
                        class_name="group-data-[state=open]:bg-[--slate-3] group-hover:bg-[--slate-3] p-2 rounded-xl transition-bg cursor-pointer",
                    ),
                ),
                class_name="group",
            ),
        ),
        rx.popover.content(
            rx.box(
                # Selection Color
                rx.box(
                    rx.text(
                        "Selection Color",
                        class_name="text-[--slate-12] font-small font-medium",
                    ),
                    rx.box(
                        *[color_item(color) for color in colors],
                        class_name="grid grid-cols-2 gap-3",
                    ),
                    class_name="flex flex-col gap-3",
                ),
                # Opacity
                rx.box(
                    rx.text(
                        f"Opacity {SettingsState.opacity[0]}",
                        class_name="text-[--slate-12] font-small font-medium",
                    ),
                    rx.slider(
                        min=0.75,
                        max=1.0,
                        step=0.025,
                        value=SettingsState.opacity,
                        size="1",
                        radius="full",
                        on_change=SettingsState.set_opacity,
                    ),
                    class_name="flex flex-col gap-3",
                ),
                class_name="border-[--slate-5] bg-[--slate-1] shadow-large border border-box rounded-xl flex flex-col gap-4 overflow-hidden px-[0.875rem] py-4",
            ),
            side="top",
            align="center",
            avoid_collisions=True,
            class_name="items-center bg-transparent !shadow-none !p-0 border-none w-[254px] overflow-visible",
        ),
    )


def textarea_prompt() -> rx.Component:
    return rx.form(
        rx.el.textarea(
            name="prompt",
            placeholder=rx.cond(
                ToolbarState.step == 2,
                "Ask me for any follow up changes...",
                "Ask me to change anything...",
            ),
            max_length=1000,
            # autofocus=True,
            enter_key_submit=True,
            disabled=ToolbarState.processing,
            class_name="border-[--slate-5] bg-[--white-1] p-[0.5rem_2.5rem_0.5rem_0.75rem] border rounded-[0.75rem] focus:ring-2 focus:ring-[--violet-9] ring-offset-0 w-full h-24 scrollbar-width-thin font-base text-[--slate-11] placeholder:text-[--slate-9] main-textarea outline-none resize-none",
        ),
        rx.cond(
            ~ToolbarState.processing,
            hint(
                text="Submit",
                content=icon_button(
                    icon="arrow-up",
                    class_name="right-3 bottom-2 !absolute",
                    style={
                        "textarea:placeholder-shown + &": {
                            "opacity": "0.65",
                            "cursor": "not-allowed",
                            # "pointer-events": "none",
                            "_hover": {
                                "background": "linear-gradient(180deg, var(--violet-9) 0%, var(--violet-10) 100%)"
                            },
                        },
                    },
                ),
            ),
            hint(
                text="Cancel",
                content=icon_button(
                    icon="square",
                    variant="destructive",
                    class_name="right-3 bottom-2 !absolute",
                ),
            ),
        ),
        # reset_on_submit=True,
        on_submit=ToolbarState.process,
        class_name="relative flex flex-col justify-end gap-3 w-full h-auto max-h-[160px]",
    )


# Bottom bar
def bottom_bar() -> rx.Component:
    from reflex_ai.playground import SettingsState

    return (
        rx.box(
            textarea_prompt(),
            rx.box(
                hint(
                    text="Hide sidebar",
                    content=rx.box(
                        rx.icon(
                            tag="panel-right-open",
                            size=20,
                            stroke_width="1.5",
                            class_name="shrink-0",
                            rotate=rx.cond(
                                SettingsState.side == "right", "180deg", "0deg"
                            ),
                        ),
                        class_name="hover:bg-[--slate-3] p-2 rounded-xl text-[--slate-10] transition-bg cursor-pointer",
                        on_click=SettingsState.setvar(
                            "is_open", ~SettingsState.is_open
                        ),
                    ),
                ),
                hint(
                    text="Select and edit",
                    content=rx.box(
                        rx.icon(
                            tag="square-mouse-pointer",
                            size=20,
                            stroke_width="1.5",
                            class_name="shrink-0",
                        ),
                        class_name="hover:bg-[--slate-3] p-2 rounded-xl text-[--slate-10] transition-bg cursor-pointer",
                    ),
                ),
                hint(
                    text=f"Move to {rx.cond(SettingsState.side == 'right', 'left', 'right')}",
                    content=rx.box(
                        rx.icon(
                            tag="arrow-left-to-line",
                            size=20,
                            stroke_width="1.5",
                            class_name="shrink-0",
                            rotate=rx.cond(
                                SettingsState.side == "right", "0deg", "180deg"
                            ),
                        ),
                        class_name="hover:bg-[--slate-3] p-2 rounded-xl text-[--slate-10] transition-bg cursor-pointer",
                        on_click=SettingsState.toggle_side,
                    ),
                ),
                rx.spacer(),
                settings_icon(),
                help_icon(),
                class_name="flex flex-row items-center gap-3 w-full",
            ),
            class_name="flex flex-col gap-3 w-full",
        ),
    )



import reflex as rx


def code_block(code: str, language: str = "python") -> rx.Component:
    return rx.box(
        rx.code_block(
            code,
            language=language,
            # wrap_long_lines=True,
            class_name="code-block",
        ),
        class_name="flex flex-col gap-3 w-full h-full overflow-hidden",
    )



import json

from pathlib import Path
import reflex as rx

from reflex_ai.utils import path_utils


class RunTimeState(rx.State):
    """Perform some actions on app startup."""

    # track the last reflex run time.
    last_reflex_run_time: rx.LocalStorage

    async def reset_scratch_dir(self, base_path):
        """Reset the scratch directory on every reflex run.

        Args:
            base_path: The root path of the client working dir.
        """
        web_path = Path(base_path).parent / ".web" / "reflex.json"

        with open(web_path) as f:
            data = json.load(f)

        last_reflex_run_time = data.get("last_reflex_run_datetime", "_")

        if not self.last_reflex_run_time == last_reflex_run_time:
            path_utils.create_scratch_dir(Path(base_path), overwrite=True)
            self.last_reflex_run_time = last_reflex_run_time



import ast
import importlib
import inspect
import os
import subprocess
import sys

from reflex_ai import paths
from reflex_ai.scripts import get_module_script


def import_module_from_scratch(module_name: str, root_dir: str, scratch_dir: str):
    """Import the module from the scratchpad.

    Args:
        module_name: The name of the module to import.
        root_dir: The root directory of the module.
        scratch_dir: The scratch directory of the module.

    Returns:
        The imported module.
    """
    module_tmp_file = os.path.join(scratch_dir, *module_name.split(".")) + ".py"
    spec = importlib.util.spec_from_file_location(
        module_name,
        module_tmp_file,
    )
    new_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(new_module)
    return new_module


def get_module_source(module_name: str) -> str:
    """Get the file path of a module in a subprocess.

    Args:
        module_name: The name of the module.

    Returns:
        The source code of the module.

    Raises:
        Exception: If the subprocess fails.
    """
    path = inspect.getfile(get_module_script)
    try:
        result = subprocess.run(
            [
                sys.executable,
                path,
                module_name,
                str(paths.base_paths[0]),
                str(paths.tmp_root_path),
            ],
            capture_output=True,
            text=True,
            check=True,
        )
    except Exception as e:
        raise Exception(e.stderr)
    # Capture the output
    module_file = result.stdout.strip().split("\n")[-1]
    with open(module_file) as f:
        return f.read()


def get_module_ast(module_name: str) -> ast.Module:
    """Get the AST of a module.

    Args:
        module_name: The name of the module.

    Returns:
        The AST of the module.
    """
    # Import the module
    module_source = get_module_source(module_name)

    # Parse the module source code into an AST
    module_ast = ast.parse(module_source)
    return module_ast


def find_node(tree, element_type: str, element_name: str) -> ast.AST | None:
    for node in ast.walk(tree):
        if (
            isinstance(node, ast.FunctionDef)
            and element_type == "function"
            and node.name == element_name
        ):
            return node
        elif (
            isinstance(node, ast.ClassDef)
            and element_type == "class"
            and node.name == element_name
        ):
            return node
        elif isinstance(node, ast.Assign) and element_type == "variable":
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id == element_name:
                    return node
        elif isinstance(node, ast.Return) and element_type == "return":
            return node
    return None



import reflex as rx
from reflex_ai.components.hint import hint
from reflex_ai.components.code_block import code_block
from reflex_ai.selection import ClickSelectionState


def selected_code_pill(text: str) -> rx.Component:
    return hint(
        text="Open in IDE",
        content=rx.link(
            rx.el.button(
                text,
                class_name="max-w-fit border-[--slate-5] bg-[--slate-3] px-3 py-[0.275rem] border rounded-[62.5rem] font-small text-[--slate-11] cursor-pointer",
            ),
            href=ClickSelectionState.selected_code_ide_hyperlink,
            is_external=True,
            class_name="w-fit",
        ),
        side="right",
    )


def selected_code_box() -> rx.Component:
    return rx.box(
        # Show the selected filename
        rx.cond(
            ClickSelectionState.selected_code_filename_and_line_range,
            selected_code_pill(
                ClickSelectionState.selected_code_filename_and_line_range
            ),
            rx.fragment(),
        ),
        # Show the selected code
        code_block(
            rx.cond(
                ClickSelectionState.code,
                ClickSelectionState.code,
                "Select a component to edit",
            )
        ),
        class_name="flex flex-col gap-3 w-full",
    )



"""Script to get the source file of a module in a sandbox."""

import builtins
import inspect
import sys

original_import = builtins.__import__


def custom_import(name, globals=None, locals=None, fromlist=(), level=0) -> object:
    """A custom import to disable reflex_ai.enable.

    Args:
        name: The name of the module to import.
        globals: The global namespace.
        locals: The local namespace.
        fromlist: The list of names to import.
        level: The level in the package hierarchy.

    Returns:
        The imported module.
    """
    if len(name) >= 3 and name[:3] == "tmp":
        name = name[3:]

    return original_import(name, globals, locals, fromlist, level)


def get_module(module_name: str, root_dir: str, scratch_dir: str):
    """Import the module from the scratchpad.

    Args:
        module_name: The name of the module to import.
        root_dir: The root directory of the module.
        scratch_dir: The scratch directory of the module.

    Returns:
        The imported module.
    """
    # Import here to avoid circular imports.
    from reflex_ai.utils.ast_utils import import_module_from_scratch

    # Import the module from the scratchpad.
    if module_name[:3] == "tmp":
        module_name = module_name[3:]
    module = import_module_from_scratch(module_name, root_dir, scratch_dir)
    module_file = inspect.getsourcefile(module)
    return module_file


if __name__ == "__main__":
    builtins.__import__ = custom_import
    module = get_module(sys.argv[1], sys.argv[2], sys.argv[3])
    builtins.__import__ = original_import
    print()
    print(module)



"""The agent that runs on the user's local client."""

import dataclasses
import difflib
import importlib
import json
import os

import httpx
from flexai import Agent
from flexai.message import ToolCall, ToolResult
from pydantic import BaseModel
from reflex.utils import console

import reflex as rx
from reflex_ai.utils import ast_utils, codegen, path_utils
from reflex_ai import paths
from reflex_cli.v2.utils import hosting


# NOTE: using BaseModel here instead of rx.Base due to FastAPI not liking the v1 models
class InternRequest(BaseModel):
    """The request to the AI intern."""

    prompt: str
    selected_code: str
    selected_module: str
    selected_function: str
    parent_component: dict[str, str]


class InternResponse(BaseModel):
    """The response from the AI intern."""

    thread_id: str
    tool_calls: list[ToolCall] = []
    final_message: str = ""


class ToolRequestResponse(BaseModel):
    """The response from the tool to the AI intern."""

    thread_id: str
    tool_results: list[ToolResult]


class EditResult(BaseModel):
    """The result of an edit."""

    thread_id: str
    diff: str
    accepted: bool


# Tools mapping to human readable names
TOOLS_HR_NAMES = {
    ast_utils.get_module_source.__name__: "Analyzing code",
    codegen.add_python_element.__name__: "Adding new elements",
    codegen.update_python_element.__name__: "Updating elements",
    codegen.delete_python_element.__name__: "Deleting elements",
}

FLEXGEN_BACKEND_URL = os.getenv("FLEXGEN_BACKEND_URL", "http://localhost:8000")


class Diff(rx.Base):
    """A diff between two files (before and after)."""

    # The name of the file
    filename: str

    # The diff of the file
    diff: str


@dataclasses.dataclass()
class LocalAgent:
    """The local agent instance."""

    # The remote backend URL.
    backend_url: str = FLEXGEN_BACKEND_URL

    # The agent instance.
    agent = Agent(
        tools=[
            ast_utils.get_module_source,
            codegen.add_python_element,
            codegen.update_python_element,
            codegen.delete_python_element,
        ],
    )

    @staticmethod
    def authorization_header(token: str) -> dict[str, str]:
        """Construct an authorization header with the specified token as bearer token.

        Args:
            token: The access token to use.

        Returns:
            The authorization header in dict format.
        """
        return {"Authorization": f"Bearer {token}"}

    async def make_request(
        self,
        endpoint: str,
        data: dict,
        timeout: int = 60,
    ) -> dict:
        """Make a request to the backend.

        Args:
            endpoint: The endpoint to send the request to.
            data: The data to send.
            timeout: The timeout for the request.

        Returns:
            The JSON response from the backend.
        """

        token, validated_info = hosting.authenticated_token()
        headers = self.authorization_header(token)

        async with httpx.AsyncClient() as client:
            resp = await client.post(
                f"{self.backend_url}/api/{endpoint}",
                data=data,
                headers=headers,
                timeout=timeout,
            )

        resp.raise_for_status()
        return resp.json()

    async def process(self, request: InternRequest):
        """Process the request from the AI intern.

        Args:
            request: The request from the AI intern.

        Yields:
            The tool use messages from the intern.
        """
        # Create the initial edit request and yield the edit id.
        response = InternResponse(
            **await self.make_request("intern", request.model_dump_json())
        )
        yield response

        # Process and yield the tool use messages.
        while True:
            tool_results: list[ToolResult] = []

            for tool_call in response.tool_calls:
                yield tool_call
                tool_result = await self.agent.invoke_tool(tool_call)
                assert tool_result.tool_call_id == tool_call.id
                tool_result = await self._handle_tool_result(
                    tool_result, tool_call, request
                )
                assert tool_result.tool_call_id == tool_call.id
                tool_results.append(tool_result)

            tool_response_request = ToolRequestResponse(
                thread_id=response.thread_id,
                tool_results=tool_results,
            )

            response = InternResponse(
                **await self.make_request(
                    "intern/tool_response", tool_response_request.model_dump_json()
                )
            )

            # Check if the response is the final response.
            yield response

    async def _handle_tool_result(
        self, tool_result: ToolResult, tool_call: ToolCall, request
    ):
        """Handle the tool response.

        Args:
            tool_result: The result from the tool.
            tool_call: The tool call item.
            request: The original request.

        Returns:
            The processed tool response.
        """
        if not tool_result.is_error and tool_call.name in [
            codegen.add_python_element.__name__,
            codegen.update_python_element.__name__,
            codegen.delete_python_element.__name__,
        ]:
            # The tool is trying to update the source code.
            # Obtain the new source code, from the diff
            suggested_ndiff = tool_result.result
            new_source = "".join(difflib.restore(suggested_ndiff, 2))
            new_source = path_utils.convert_from_scratch_code(
                paths.base_paths[0].name, new_source
            )

            # Obtain the file that the modified module lives in (within the main directory)
            module_modified = "tmp" + tool_call.input["module_name"]
            modified_filename = importlib.util.find_spec(module_modified).origin
            modified_filename = modified_filename.replace(
                str(paths.tmp_root_path), str(paths.base_paths[0].parent)
            )
            try:
                # Obtain the validation function (i.e. index) and the file (in the main directory) where it lives
                main_directory_validation_page = request.parent_component["function"]
                main_directory_validation_file = request.parent_component[
                    "filename"
                ].replace(str(paths.tmp_root_path), str(paths.base_paths[0].parent))
                # Validate the new source
                path_utils.validate_source(
                    new_source,
                    modified_filename,
                    main_directory_validation_page,
                    main_directory_validation_file,
                )
            except Exception as e:
                # This code has failed: return the error message
                tool_result = ToolResult(
                    tool_call_id=tool_result.tool_call_id,
                    result=e.stderr,
                    execution_time=0,
                    is_error=True,
                )
            else:
                # This code has passed: write it into the correct file within the * scratch directory *
                write_to = modified_filename.replace(
                    str(paths.base_paths[0].parent), str(paths.tmp_root_path)
                )
                with open(write_to, "w") as file:
                    new_source = path_utils.convert_to_scratch_code(
                        paths.base_paths[0].name, new_source
                    )
                    content = path_utils.format_code(new_source)
                    file.write(content)
                tool_result = ToolResult(
                    tool_call_id=tool_result.tool_call_id,
                    result=codegen.to_unified_diff(suggested_ndiff),
                    execution_time=tool_result.execution_time,
                )

        console.info("\n\n\n")
        console.info(tool_result)
        console.info("\n\n\n")
        return tool_result

    async def confirm_change(self, thread_id: str, diffs: list[Diff], accepted: bool):
        """Accept or reject the change.

        Args:
            thread_id: The thread ID.
            diffs: The diffs to accept or reject.
            accepted: Whether the change is accepted or rejected.
        """
        console.info(f"Confirming change for {thread_id} (accepted: {accepted})")
        await self.make_request(
            "intern/edit_result",
            data=EditResult(
                thread_id=thread_id,
                diff=json.dumps([d.dict() for d in diffs]),
                accepted=accepted,
            ).model_dump_json(),
        )

        if accepted:
            path_utils.commit_scratch_dir(
                paths.base_paths[0], [d.filename for d in diffs]
            )
        else:
            path_utils.create_scratch_dir(paths.base_paths[0], overwrite=True)



import reflex as rx
from reflex_ai.toolbar import ToolbarState


def tl_item(text: str, index: int) -> rx.Component:
    return rx.cond(
        ToolbarState.processing & (index == ToolbarState.tools_used.length() - 1),
        rx.el.li(
            rx.box(
                rx.spinner(size="1"),
                class_name="left-[-0.65rem] absolute flex justify-center items-center bg-[--slate-4] p-[0.225rem] rounded-full w-5 h-5",
            ),
            text,
            class_name="text-[--slate-11]",
        ),
        rx.el.li(
            rx.box(
                rx.icon(tag="check", size=13),
                class_name="left-[-0.65rem] absolute flex justify-center items-center bg-[--green-4] p-[0.225rem] rounded-full w-5 h-5",
            ),
            text,
            class_name="text-[--green-10]",
        ),
    )


def process_box() -> rx.Component:
    return rx.box(
        rx.box(
            rx.el.ol(
                rx.foreach(
                    ToolbarState.tools_used, lambda text, index: tl_item(text, index)
                ),
                class_name="relative flex flex-col gap-4 border-[--green-5] p-0 pl-6 border-l-[1.5px] font-small text-[--slate-11] list-none",
            ),
            class_name="ps-4",
        ),
        class_name="flex flex-col w-full",
    )



import reflex as rx


def get_main_styles():
    return rx.el.style(
        """
@tailwind base;
@tailwind components;
@tailwind utilities;

@import url('https://fonts.googleapis.com/css2?family=Instrument+Sans:wght@400;500;600;700;800&display=swap');
@import url('https://fonts.googleapis.com/css2?family=Source+Code+Pro:wght@400;500&display=swap');

:root {
    --font-instrument-sans: 'Instrument Sans', sans-serif;
    --font-source-code-pro: 'Source Code Pro', monospace;
    --shadow-small: 0px 2px 4px 0px rgba(28, 32, 36, 0.05);
    --shadow-medium: 0px 4px 8px 0px rgba(28, 32, 36, 0.04);
    --shadow-large: 0px 24px 12px 0px rgba(28, 32, 36, 0.02), 0px 8px 8px 0px rgba(28, 32, 36, 0.02), 0px 2px 6px 0px rgba(28, 32, 36, 0.02);
    .font-small {
        font-family: var(--font-instrument-sans);
        font-size: 0.9rem;
        font-style: normal;
        font-weight: 500;
        line-height: 1.25rem;
        letter-spacing: -0.01094rem;
    }
    .font-md {
        font-family: var(--font-instrument-sans);
        font-size: 1.125rem;
        font-style: normal;
        font-weight: 500;
        line-height: 1.625rem;
        letter-spacing: -0.01688rem;
    }
    .font-smbold {
        font-family: var(--font-instrument-sans);
        font-size: 1rem;
        font-style: normal;
        font-weight: 600;
        line-height: 1.5rem;
        letter-spacing: -0.015rem;
    }
    .font-base {
        font-family: var(--font-instrument-sans);
        font-size: 1rem;
        font-style: normal;
        font-weight: 500;
        line-height: 1.5rem;
        letter-spacing: -0.015rem;
    }
    .font-large {
        font-family: var(--font-instrument-sans);
        font-size: 1.5rem;
        font-weight: 600;
        line-height: 2rem;
        letter-spacing: -0.03rem;
    }
    .font-x-large {
        font-family: var(--font-instrument-sans);
        font-size: 2rem;
        font-style: normal;
        font-weight: 600;
        line-height: 2.5rem;
        letter-spacing: -0.06rem;
    }
    .font-xxx-large {
        font-family: var(--font-instrument-sans);
        font-size: 3.5rem;
        font-style: normal;
        font-weight: 600;
        line-height: 4rem;
        letter-spacing: -0.175rem;
    }
    .font-code {
        font-family: var(--font-source-code-pro);
        font-size: 0.875rem;
        font-style: normal;
        font-weight: 400;
        line-height: 1.5;
        letter-spacing: -0.01313rem;
    }
    .code-block {
        width: 100% !important;
        height: auto !important;
        margin: 0 !important;
        padding: 1rem !important;
        max-height: 450px !important;
        border-radius: 0.75rem !important;
        border: 1px solid var(--slate-5) !important;
        background: var(--white-1) !important;
        color: var(--slate-12) !important;
        resize: none !important;
        outline: none !important;
        scrollbar-width: thin !important;
        font-family: var(--font-source-code-pro) !important;
    }
    .code-block > * {
        background: transparent !important;
        font-family: var(--font-source-code-pro) !important;
        font-size: 0.875rem !important;
        font-style: normal !important;
        font-weight: 400 !important;
        line-height: 1.5 !important;
        letter-spacing: -0.01313rem !important;
    }
    .scrollbar-width-thin {
        scrollbar-width: thin;
    }
    .transition-color {
        transition: color 0.035s ease-out;
    }
    .transition-bg {
        transition: background-color 0.075s ease-out;
    }
}

@layer components {
    .font-instrument-sans {
        font-family: var(--font-instrument-sans);
    }
    .font-source-code-pro {
        font-family: var(--font-source-code-pro);
    }
    .shadow-large {
        box-shadow: var(--shadow-large);
    }
    .shadow-medium {
        box-shadow: var(--shadow-medium);
    }
    .shadow-small {
        box-shadow: var(--shadow-small);
    }
}

.light,
.light-theme {
	--white-1: #ffffff;
}

.dark,
.dark-theme {
	--white-1: #0E0F10;
}
"""
    )



"""The local AI toolbar to hook up with the server-side reflex agent."""

import reflex as rx
from reflex.utils import console
from reflex_ai.selection import ClickSelectionState
from reflex_ai.local_agent import (
    LocalAgent,
    InternResponse,
    InternRequest,
    Diff,
    TOOLS_HR_NAMES,
)
from reflex_ai.utils import path_utils


class ToolbarState(rx.State):
    """The toolbar state."""

    tools_used: list[str] = ["Preparing request"]
    current_step: str = "selected_code"  # The current step name of the process (selected_code, processing, review_changes)
    processing: bool = False
    selected_id: str = ""
    code: str = ""
    prompt: str = ""
    step: int = 0
    diff: list[Diff] = []
    selected_diff: Diff = Diff(filename="", diff="")  # The selected diff
    changes_comment: str = ""
    thread_id: str = ""
    _agent: LocalAgent = LocalAgent()

    @rx.background
    async def process(self, prompt: dict[str, str]):
        """Process the user's prompt.

        Args:
            prompt: The prompt from the user from the form input.

        Yields:
            Updates to the frontend.
        """
        # Set the processing flag to True.
        async with self:
            self.thread_id = ""
            self.start_processing()
            self.prompt = prompt["prompt"]
            selection_state = await self.get_state(ClickSelectionState)
        yield

        # Create the intern request.
        request = InternRequest(
            prompt=prompt["prompt"],
            selected_code=selection_state.code,
            selected_module=selection_state.selection.get_selection_module(),
            selected_function=selection_state.selection.function_name,
            parent_component=selection_state.selection.modules[-1],
        )

        # Process the messages with the local agent.
        async for response in self._agent.process(request):
            console.info(response)
            async with self:
                # Check if it's an InternResponse.
                if isinstance(response, InternResponse):
                    self.thread_id = response.thread_id
                    if response.final_message:
                        # TODO: set the final message in the state.
                        break

                # It's a tool use message.
                else:
                    self.tools_used.append(
                        TOOLS_HR_NAMES.get(response.name, response.name)
                    )
            yield

        async with self:
            self.finish_processing()
        yield
        # Trigger a hot reload.
        self.trigger_reload()

    def start_processing(self):
        self.tools_used = [self.tools_used[0]]
        self.processing = True
        self.current_step = "processing"
        self.step = 1

    def finish_processing(self):
        self.selected_diff = self.diff[0]
        # self.changes_comment = json.dumps(self.messages[-1].content)
        self.current_step = "review_changes"
        self.step = 2
        self.processing = False

    async def confirm_change(self, accepted: bool):
        """Accept the current diff.

        Args:
            accepted: Whether the diff is accepted or not.
        """
        thread_id, diff = self.thread_id, self.diff
        self.reset()
        await self._agent.confirm_change(thread_id, diff, accepted)
        if not accepted:
            self.trigger_reload()

    def trigger_reload(self):
        """Trigger a hot reload."""
        config = rx.config.get_config()
        app_name = config.app_name
        filename = f"{app_name}/{app_name}.py"
        contents = open(filename).read()
        with open(filename, "w") as f:
            f.write(contents)

    def load_diff(self):
        diff = path_utils.directory_diff()
        self.diff = [
            Diff(filename=str(filename), diff="\n".join(diff))
            for filename, diff in diff.items()
            if (len(diff) > 0 or ("__init__.py" in filename))
        ]



import reflex as rx


def hint(
    text: str,
    content: rx.Component,
    side: str = "top",
    active: bool = False,
    class_name: str = "",
    **props,
) -> rx.Component:
    return rx.hover_card.root(
        rx.hover_card.trigger(content),
        rx.hover_card.content(
            rx.text(text),
            side=side,
            align="center",
            class_name="flex justify-center items-center bg-[--slate-11] px-1.5 py-0.5 rounded-lg font-medium text-sm text-[--slate-1]",
        ),
        class_name=class_name,
        default_open=active,
        open_delay=80,
        close_delay=80,
        **props,
    )



import ast
import difflib
import astor
from reflex_ai.utils import ast_utils, path_utils


def get_ast_and_source(module_name: str) -> tuple[ast.Module, str]:
    """Get the AST and source code of a module.

    Args:
        module_name: The name of the module.

    Returns:
        The AST and source code of the module.
    """
    source = ast_utils.get_module_source(module_name)
    tree = ast.parse(source)
    return tree, source


def find_target_node(
    tree: ast.Module, element_type: str, element_name: str, class_name: str = None
) -> ast.AST:
    """Find the target node in the AST.

    Args:
        tree: The AST of the module.
        element_type: The type of element to find (function, class, or variable).
        element_name: The name of the element to find.
        class_name: The name of the class to search within (if applicable).

    Returns:
        The target node in the AST.

    Raises:
        ValueError: If the class or element is not found.
    """
    if class_name:
        class_node = ast_utils.find_node(tree, "class", class_name)
        if not class_node:
            raise ValueError(f"Class {class_name} not found")
        return ast_utils.find_node(class_node, element_type, element_name)
    return ast_utils.find_node(tree, element_type, element_name)


def generate_diff(previous_source: str, new_tree: ast.Module) -> list[str]:
    """Generate the diff between the previous source and the new tree.

    Args:
        previous_source: The previous source code.
        new_tree: The new AST.

    Returns:
        The diff between the previous source and the new tree.
    """
    new_source = astor.to_source(new_tree)
    return list(difflib.ndiff(previous_source, new_source))


def to_unified_diff(ndiff: str) -> str:
    """Convert a list of diff lines to a unified diff string.

    Args:
        ndiff: The list of diff lines.

    Returns:
        The unified diff string.
    """
    app_name = path_utils.get_app_name()
    previous_source = path_utils.format_code(
        path_utils.convert_from_scratch_code(
            app_name, "".join(difflib.restore(ndiff, 1))
        )
    )
    new_source = path_utils.format_code(
        path_utils.convert_from_scratch_code(
            app_name, "".join(difflib.restore(ndiff, 2))
        )
    )
    return "\n".join(
        difflib.unified_diff(
            previous_source.splitlines(),
            new_source.splitlines(),
        )
    )


def add_python_element(
    module_name: str, content: str, class_name: str = None
) -> list[str]:
    """Add a new element (class, function, or variable) to a Python file after the imports.

    Args:
        module_name: The name of the module.
        content: The content of the new element.
        class_name: The name of the class to add the element to (if applicable).

    Returns:
        The diff between the previous source and the new source.

    Raises:
        ValueError: If the class is not found, or if we try to add in a duplicate element
    """
    tree, previous_source = get_ast_and_source(module_name)
    new_element = ast.parse(content).body[0]

    if isinstance(new_element, ast.FunctionDef) or isinstance(
        new_element, ast.ClassDef
    ):
        element_name = new_element.name
    elif isinstance(new_element, ast.Assign):
        for target in new_element.targets:
            if isinstance(target, ast.Name):
                element_name = target.id
    else:
        raise ValueError(
            "The new element you want to create is not a class, function, or variable."
        )

    if find_target_node(tree, "class", element_name, class_name):
        raise ValueError(f"Class {element_name} already exists")

    if find_target_node(tree, "function", element_name, class_name):
        raise ValueError(f"Function {element_name} already exists")

    if find_target_node(tree, "variable", element_name, class_name):
        raise ValueError(f"Variable {element_name} already exists")

    if class_name:
        class_node = ast_utils.find_node(tree, "class", class_name)
        if not class_node:
            raise ValueError(f"Class {class_name} not found")
        target_body = class_node.body
    else:
        target_body = tree.body

    # Find the position after the last import statement
    insert_position = 0
    for i, node in enumerate(target_body):
        if isinstance(node, (ast.Import, ast.ImportFrom)):
            insert_position = i + 1
        else:
            break
    target_body.insert(insert_position, new_element)

    return generate_diff(previous_source, tree)


def update_python_element(
    module_name: str,
    element_type: str,
    element_name: str,
    new_content: str,
    class_name: str = None,
) -> list[str]:
    """Update the source code of an element (class, function, or variable) in a module.

    Args:
        module_name: The name of the module.
        element_type: The type of element to update (class, function, or variable).
        element_name: The name of the element to update.
        new_content: The new content of the element.
        class_name: The name of the class containing the element (if applicable).

    Returns:
        The diff between the previous source and the new source.

    Raises:
        ValueError: If the element is not found.
    """
    if element_type not in ["class", "function", "variable"]:
        raise ValueError(
            f"Invalid element type {element_type} - must be one of 'class', 'function', or 'variable'"
        )

    tree, previous_source = get_ast_and_source(module_name)
    target_node = find_target_node(tree, element_type, element_name, class_name)

    if not target_node:
        raise ValueError(f"{element_type.capitalize()} {element_name} not found")

    new_node = ast.parse(new_content).body[0]

    if element_type == "variable":
        target_node.value = new_node.value
    elif element_type == "function":
        # Update the function signature
        target_node.name = new_node.name
        target_node.args = new_node.args
        target_node.returns = new_node.returns
        # Preserve docstring if present
        for item in target_node.body:
            if isinstance(item, ast.Expr) and isinstance(item.value, ast.Str):
                new_node.body.insert(0, item)
                break
        target_node.body = new_node.body
    else:
        target_node.body = new_node.body

    return generate_diff(previous_source, tree)


def delete_python_element(
    module_name: str, element_type: str, element_name: str, class_name: str = None
) -> list[str]:
    """Delete an element (class, function, or variable) from a module.

    Args:
        module_name: The name of the module.
        element_type: The type of element to delete (class, function, or variable).
        element_name: The name of the element to delete.
        class_name: The name of the class containing the element (if applicable).

    Returns:
        The diff between the previous source and the new source.

    Raises:
        ValueError: If the class or element is not found.
    """
    tree, previous_source = get_ast_and_source(module_name)

    if class_name:
        class_node = ast_utils.find_node(tree, "class", class_name)
        if not class_node:
            raise ValueError(f"Class {class_name} not found")
        class_node.body = [
            node
            for node in class_node.body
            if not (
                (
                    isinstance(node, ast.FunctionDef)
                    and element_type == "function"
                    and node.name == element_name
                )
                or (
                    isinstance(node, ast.Assign)
                    and element_type == "variable"
                    and any(
                        t.id == element_name
                        for t in node.targets
                        if isinstance(t, ast.Name)
                    )
                )
            )
        ]
    else:
        tree.body = [
            node
            for node in tree.body
            if not (
                (
                    isinstance(node, ast.FunctionDef)
                    and element_type == "function"
                    and node.name == element_name
                )
                or (
                    isinstance(node, ast.ClassDef)
                    and element_type == "class"
                    and node.name == element_name
                )
                or (
                    isinstance(node, ast.Assign)
                    and element_type == "variable"
                    and any(
                        t.id == element_name
                        for t in node.targets
                        if isinstance(t, ast.Name)
                    )
                )
            )
        ]

    return generate_diff(previous_source, tree)
